/*
  Usage: set environment variable QUERY_STRING to some combination of:

	p=planet-name[+planet-name...]  // list of planet names
	d=yyyy-mm-dd[Thh:mm]            // UT date with optional following time

  Then run the program.  Its output is a web service header producing plain
  text with a max cache life of 1 miute.  The payload of this is a 1-line status
  followed by N other lines.  Every line is terminated by "\r\n".

  If the status is "200", then the following lines are in CSV format, with each
  line having 4 fields:
   - planet abbreviation: first 2 chars of the planet name, in upper case.  See
     the array Planets[] below to remember what those are.
   - longitude: "deg SIGN min sec", with SIGN the first 2 chars of the sign name
     (AR = Aries, TA = Taurus, etc.)
   - latitude: "deg:min:sec" north/south of the equator.  Here, a negative
     number of degrees indicates south.
   - distance: the plaent's distance from the earth, in AU
   - speed long: "deg:min:sec" per hour of its perceived traversal speed.
     Here, a negative number of degrees indicates a retrograde motion.

  If specified, planet-name is a list of planet names separated by '+'.  The
  known list of planet names is given in Planets[] below.  In that list, "sol"
  is an alias for "sun" and "luna" is an alias for "moon".  Case is irrelevant.

  If specified, time is a Universal Time (UT) date plus optional time.  The date
  is YYYY-MM-DD; no other format is accepted. If included, the time is prefixed
  by capital 'T' and is specified as HH:MM; it can include ":SS" but those are
  ignored.

  This file uses functions made available through Swiss Ephemeris. Authors of
  the Swiss Ephemeris: Dieter Koch and Alois Treindl.  Swiss Ephemeris is
  copyright (C) 1997 - 2008 Astrodienst AG, Switzerland.  All rights reserved.

  The authors of Swiss Ephemeris have no control or influence over any of
  the derived works, i.e. over software or services created by other
  programmers which use Swiss Ephemeris functions.

  The names of the authors or of the copyright holder (Astrodienst) must not
  be used for promoting any software, product or service which uses or contains
  the Swiss Ephemeris. This copyright notice is the ONLY place where the
  names of the authors can legally appear, except in cases where they have
  given special permission in writing.

  The trademarks 'Swiss Ephemeris' and 'Swiss Ephemeris inside' may be used
  for promoting such software, products or services.
*/

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "strings.h"
#include "ctype.h"
#include <errno.h>
#include <math.h>
#include "swephexp.h"      /* this includes  "sweodef.h" */
#include <time.h>

#define DEBUG 0

#if DEBUG
#define DBG_OUT(a) printf a;
static char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun"
						 ,"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
#else
#define DBG_OUT(a)
#endif

#define ARRAY_LEN(a) (sizeof(a) / sizeof((a)[0]))

#define MAKE_QUOTED_ENTRY(ptr,size,len,fn,arg) \
	do {	\
		*ptr++ = '"'; \
		size--; \
		len = fn(arg,ptr,size); \
		ptr += len; \
		size -= len; \
		*ptr++ = '"'; \
		*ptr++ = ','; \
		size -= 2; \
	} while (0)

typedef struct {
	char		*dsp;	// date string pointer
	int			*plist;	// planet number list;
	time_t		 now;	// date encoded for use by gmtime()
	struct tm	 time;	// buffer for use by gmtime()
	int			 pct;	// number of planets in plist
} EINFO;

static struct {
	char	 name[8];
	char	 abbr[4];
	int		 id;
} Planets[] = {
	 { "sun"    , "SU", SE_SUN    }
	,{ "sol"    , "SU", SE_SUN    }
	,{ "moon"   , "MO", SE_MOON   }
	,{ "luna"   , "MO", SE_MOON   }
	,{ "mercury", "ME", SE_MERCURY}
	,{ "venus"  , "VE", SE_VENUS  }
	,{ "mars"   , "MA", SE_MARS   }
	,{ "jupiter", "JU", SE_JUPITER}
	,{ "saturn" , "SA", SE_SATURN }
	,{ "uranus" , "UR", SE_URANUS }
	,{ "neptune", "NE", SE_NEPTUNE}
	,{ "pluto"  , "PL", SE_PLUTO  }
};

static int
dbl_to_dms(double x, char *buf, size_t bufsize)
{
	int	hrs;
	char deg[8];

	sprintf(deg, "%.0lf", (x < 0.0) ? ceil(x) : floor(x));
	x -= (double)(int) x;
	if (x < 0.0) x = -x;
	x *= 60.0;
	hrs = (int) x;
	x -= (double) hrs;
	return snprintf(buf, bufsize, "%s:%d:%.2lf", deg, hrs, x * 60.0);
}

static int
get_planet_id(char *pp, int i)
{
	int	j;

	for (j = 0; j < (int) ARRAY_LEN(Planets); j++)
	{
		if ((strlen(Planets[j].name) == i) && !strncasecmp(Planets[j].name, pp, i))
			return Planets[j].id;
	}
	return -1;
}

static void
get_planet_list(char *qs, EINFO *ep)
{
	char	*pp = strstr(qs, "p=");
	int		 i;
	int		 id;
	int		 j;

	if (pp)
	{
		j = 0;
		pp += 2;
A:		for (i = 0; isalpha(pp[i]); i++) { }
		if ((i > 0) && ((id = get_planet_id(pp, i)) >= 0))
		{
			DBG_OUT(("pp : '%.*s'=%d\n", i, pp, id))
			ep->plist[j++] = id;
			if (j < ep->pct)
			{
				pp += i;
			    if (*pp++ == '+') goto A;
			}
		}
		ep->pct = j;
	}
	return;
}

static void
get_date(char *qs, EINFO *ep)
{
	char	*tp = strstr(qs, "d=");
	time_t	 now;
	int		 year;
	int		 month;
	int		 day;
	int		 hour;
	int		 min;
	int		 c;
	static int dayno[] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };

	if (tp)
	{
		year = month = day = hour = min = 0;
		tp += 2;

		// Years are  valid for 1970 .. 2099
		while (isdigit(c = *tp++)) { year = (year * 10) + (c - '0'); }
		if ((year < 1970) || (year >= 2100) || (c != '-'))
		{
			DBG_OUT(("Bad date: %d-\n", year))
			return;
		}

		// Months are valid for 1 .. 12
		while (isdigit(c = *tp++)) { month = (month * 10) + (c - '0'); }
		if ((month < 1) || (month > 12) || (c != '-'))
		{
			DBG_OUT(("Bad date: %d-%02d-\n", year, month))
			return;
		}

		/*
		 * Days are valid for 1 .. {28 .. 31} depending on the month/year.
		 * Here, we cheat: 2000 was a leap year, so every 4 years is leap
		 * for the calendar we recognize.
		 */
		while (isdigit(c = *tp++)) { day = (day * 10) + (c - '0'); }
		if ((day < 1)
		 || (day > ((month == 2)
						? (28 + !(year & 3))
						: (30 + ((month & 1) ^ (month > 7))))))
		{
			DBG_OUT(("Bad date: %d-%02d-%02d\n", year, month, day))
			return;
		}

		/* The date is valid. Do we have a time? */
		if (c == 'T')
		{
				while (isdigit(c = *tp++)) { hour = (hour * 10) + (c - '0'); }
				if ((hour > 23) || (c != ':'))
					return;
				while (isdigit(c = *tp++)) { min = (min * 10) + (c - '0'); }
				if ((min > 59) || (c && (c != ':')))
					return;
		}
		else if (c && (c != '&'))
		{
			DBG_OUT(("Bad terminal: %c%s\n", c, tp));
			return;
		}

		DBG_OUT(("OK date: %d-%02d-%02d %02d:%02d\n", year, month, day, hour, min))

		// Get a count of leap years in our date
		c = (year - 1968) >> 2;
		if (!(year & 3))
			c--;

		// "year" becomes years-since-1970 + number-of-leap-years
		now = (time_t) (year - 1970 + c) * (time_t) (365 * 24 * 60 * 60);

		// "month" now moves to the range 0 .. 11
		c = dayno[ month - 1 ];
		if ((month > 2) && !(year & 3))
			c++;
		now += (time_t) (c * 24 * 60 * 60);

		// Add the days
		now += (time_t) ((day - 1) * 24 * 60 * 60);

		// Add the hours
		now += (time_t) (hour * 60 * 60);

		// Add the minutes
		now += (time_t) (min * 60);

		ep->now = now;
	}

	return;
}

static char *
get_planet_abbr(int id)
{
	int	i;

	for (i = 0; i < (int) ARRAY_LEN(Planets); i++)
	{
		if (Planets[i].id == id)
			return Planets[i].abbr;
	}
	return "??";
}

static int
dbl_to_sign(double x, char *buf, size_t buflen)
{
	int i;
	int deg;
	int hrs;
	int sec;
	// Signs are 30 degrees each, beginning with Aries at 0deg - 29.9999 deg
	static char* sign[] = { "AR", "TA", "GE", "CA", "LE", "VI", "LI", "SC", "SA", "CA", "AQ", "PI" };

	deg = (int) x;
	x = (x - (double) deg) * 60.0;
	i = deg / 30;
	deg -= (i * 30);

	hrs = (int) x;
	sec = (int) ((x - (double) hrs) * 60.0);

	return snprintf(buf, buflen, "%d %s %d %d", deg, sign[i], hrs, sec);
}

int main(int argc, char *argv[])
{
	char	*sp, sdate[AS_MAXCH], snam[40], serr[AS_MAXCH]; 
	char	*qs;	// query string
	char	*when;	// requested date string
	char	*pname;	// abbreviated planet name
	char	*outbp;
	size_t	 outblen;
	size_t	 outbsize;
	size_t	 outbtot;
	int		 jday, jmon, jyear;
	double	 jut = 0.0;
	double	 tjd_ut, te, x2[6];
	long	 iflag, iflgret;
	int		 p;
	EINFO	 einfo;
	int		 plist[ ARRAY_LEN(Planets) ];
	char	 outbuf[ARRAY_LEN(Planets) ][ 128 ];
	static char Status[] = "200\r\n";

	einfo.now   = 0;
	einfo.plist = plist;
	p = 0;
	do {
		einfo.plist[p++] = Planets[p].id;
	} while (p < ARRAY_LEN(Planets));
	einfo.pct = p;

	if ((qs = ((argc > 1) ? argv[1] : getenv("QUERY_STRING"))))
	{
//		fprintf(stderr, "ephemeris_data query string: %s\n", qs);
		get_planet_list(qs, &einfo);
		get_date(qs, &einfo);
	}

	/* Use now as the default time if none was specified */
	if (einfo.now == 0)
		time(&einfo.now);

	/* Convert time_t into date parts */
	einfo.time = *gmtime(&einfo.now);
	jday = einfo.time.tm_mday;
	jmon = einfo.time.tm_mon + 1;
	jyear = einfo.time.tm_year + 1900;
	DBG_OUT(("date (dmy): %d.%d.%d\n", jday, jmon, jyear))

	/* Convert HH:MM:SS to a floating-point double */
	jut = (double) einfo.time.tm_hour
		+ ((double) einfo.time.tm_min / 60.0)
		+ ((double) einfo.time.tm_sec / 3600.0);

	 /* we have day, month and year and convert to Julian day number */
	tjd_ut = swe_julday(jyear,jmon,jday,jut,SE_GREG_CAL);       

#if 1
	printf("%d-%02d-%02dT%02d:%02d:%02d == %lf\n",
			jyear, jmon, jday,
			einfo.time.tm_hour, einfo.time.tm_min, einfo.time.tm_sec,
			tjd_ut);
	exit(0);
#endif

	/* Adjust the date for hour and minute */
	tjd_ut += (double)((einfo.time.tm_hour * 60 * 60) + (einfo.time.tm_min * 60)) / (24.0 * 60.0 * 60.0);

	/*
	 * compute Ephemeris time from Universal time by adding delta_t
	 * not required for Swisseph versions smaller than 1.60
	 */
	/* te = tjd_ut + swe_deltat(tjd_ut); */

	DBG_OUT(("date: %02d.%s.%d at %02d:%02d Universal time\n",
			 jday, Months[jmon - 1], jyear,
			 einfo.time.tm_hour, einfo.time.tm_min))
	DBG_OUT(("planet,longitude,latitude,distance,\"speed long.\"\n"))

	iflag = SEFLG_SPEED;

	/*
	 * a loop over all requested planets
	 */
	outbtot = 0;
	for (p = 0; p < einfo.pct; p++) {
		/*
		 * do the coordinate calculation for this planet p
		 * if there is a problem, a negative value is returned and an
		 * error message is in serr.
		 */
		iflgret = swe_calc_ut(tjd_ut, einfo.plist[p], iflag, x2, serr);
		if (iflgret < 0)
			 printf("error: %s\n", serr);

		// Set up this buffer to receive data
		outbsize = sizeof (outbuf[0]);
		outbp = outbuf[p];

		// get the name of the planet p
		// swe_get_planet_name(einfo.plist[p], snam);

		// Place planet name in the buffer
		strcpy(outbp, get_planet_abbr(einfo.plist[p]));
		outblen = strlen(outbp);
		outbp += outblen;
		outbsize -= outblen;
		*outbp++ = ',';
		outbsize--;

		// Place current planet position in the buffer
		MAKE_QUOTED_ENTRY(outbp, outbsize, outblen, dbl_to_sign, x2[0]);

		// Place current planet latitude in the buffer
		MAKE_QUOTED_ENTRY(outbp, outbsize, outblen, dbl_to_dms, x2[1]);

		// Place current planet distance in the buffer
		sprintf(outbp, "%lf", x2[2]);
		outblen = strlen(outbp);
		outbsize -= outblen;
		outbp += outblen;
		*outbp++ = ',';
		outbsize--;

		// Place current planet speed longitude in the buffer
		MAKE_QUOTED_ENTRY(outbp, outbsize, outblen, dbl_to_dms, x2[3]);

		// Finish the line
		outbp[-1] = '\r';
		outbp[0]  = '\n';
		outbp[1]  = '\0';
		outbsize--;

		// Keep track of the total size
		outbtot += (sizeof(outbuf[0]) - outbsize);
	}

	/* Print what we came up with */
	outbtot += sizeof (Status) - 1;
	printf("Content-Type: text/plain\r\nCache-Control: max-age=60\r\nContent-Length: %d\r\n\r\n%s",
		   (int) outbtot, Status);
	for (p = 0; p < einfo.pct; p++)
	{
		printf("%s", outbuf[p]);
	}

	return OK;
}
